#!BPY

# """
# Name: 'Curve Pencil Tool'
# Blender: 232
# Group: 'Misc'
# Submenu: 'Paint on Floor' Floor
# Submenu: 'Paint on Near-Clipping' NearClipping
# Submenu: 'Paint on Far-Clipping' FarClipping
# Tooltip: 'Create curves by painting strokes'
# """

__author__ = "A. Humeniuk"
__url__ = "none"
__version__ = "0.1 01/2006"
__email__ = "Mr. Humeniuk, at_carrier_pigeons:web*de", "pencil_curve"
__bpydoc__ = """\

Shortcuts:
	Q:	Quit the script
	X:	Clear the canvas
	M:	Create Bezier Curves from the Strokes drawn so far on the canvas
"""

import Blender
from Blender.BGL import *
from Blender.Curve import *
from Blender.Draw import *
from Blender.Mathutils import *
import math

argv = __script__['arg']
print argv

sc = Blender.Scene.GetCurrent()

NO_STROKE =	0
NEW_STROKE =	1
DO_STROKE =	2
END_STROKE =	3

stroke_flag = NO_STROKE

tolerance_angle = 10
min_point_dist = 0.01


def distance2D(p1, p2):
	dx = p2[0] - p1[0]
	dy = p2[1] - p1[1]
	dist = math.sqrt(dx*dx + dy*dy)
	return(dist)

def distancePointToLine(p, linetp):
	v1 = Vector(linetp[0][0] - p[0], linetp[0][1] - p[1])
	v2 = Vector(linetp[1][0] - p[0], linetp[1][1] - p[1])
	angle = AngleBetweenVecs(v1, v2)
	return(angle)

def nextDeviatingPoint(ps, start, tolerance):
	i = start
	while(i < len(ps) - 6 ):
		line = ps[start], ps[i+1]
		if (distancePointToLine(ps[i+2], line) > tolerance):
			return(i+1)
		i += 1
	return(i+1)	# never remove the last two triples


class Stroke2D:
	def __init__(self):
		self.min_p_dist = min_point_dist
		self.color = [1.0, 0.2, 0.2]
		self.points = []
	def add_point(self, x, y):
		p = [x, y]
		# append the point only if its far away from the previous
		plen = len(self.points)
		if (plen > 0):
			p_prev = self.points[plen - 1]
			if (distance2D(p, p_prev) < self.min_p_dist):
				return
		self.points.append(p)
		Redraw(0)
	def simplify(self):
		ps = self.points[0:]
		last_i = len(ps) - 2
		simple_ps = []
		i = 0
		while(i < last_i):
			simple_ps.append(ps[i])
			i = nextDeviatingPoint(ps, i, tolerance_angle)
		# replace points by simplifies points
		del self.points[0:]
		self.points = simple_ps[0:]
		Redraw(0)
	def draw(self):
		glColor3f(self.color[0], self.color[1], self.color[2])
		glBegin(GL_LINE_STRIP);
		for p in self.points:
			(xi, yi) = canvas.normCoordsToAreaCoords(p[0], p[1])
			glVertex2f(xi, yi)
  		glEnd();
	def convertToNurb(self, curve, nr):
		ps = self.points[0:]
		# first point
		hd = ps[1][0]-ps[0][0], ps[1][1]-ps[0][1]
		btuple = ps[0][0]-hd[0], ps[0][1]-hd[1], 0.0, ps[0][0], ps[0][1], 0.0, ps[1][0], ps[1][1], 0.0
		bezt = Blender.BezTriple.New(btuple)
		curve.appendNurb(bezt)
		# other points
		# If number of points is not a multiple of 3 the remaining points are ignored */
		nr_bezt = len(ps) / 3
		if (nr_bezt < 2):
			# add a second triple so that a curve is visible
			btuple = ps[0][0], ps[0][1], 0.0, ps[1][0], ps[1][1], 0.0, ps[1][0]+hd[0], ps[1][1]+hd[1], 0.0
			bezt = Blender.BezTriple.New(btuple)
			curve.appendPoint(nr, bezt)
		for i in range(1, nr_bezt):
			btuple = ps[i*3-1][0], ps[i*3-1][1], 0.0, ps[i*3][0], ps[i*3][1], 0.0, ps[i*3+1][0], ps[i*3+1][1], 0.0
			bezt = Blender.BezTriple.New(btuple)
			curve.appendPoint(nr, bezt)

class Curve2D:
	def __init__(self):
		self.strokes = []
		self.curstroke = Stroke2D()
	def add_stroke(self):
		new_stroke = Stroke2D()
		self.strokes.append(new_stroke)
	def paint_stroke(self, x, y):
		# the last stroke is the current
		self.curstroke = self.strokes[len(self.strokes) - 1]
		self.curstroke.add_point(x, y)
	def finish_stroke(self):
		self.curstroke.simplify()
	def draw(self):
		for stroke in self.strokes:
			stroke.draw()
	def delete_strokes(self):
		del self.strokes[0:]
	def make_real(self):
		#### Create a Bezier Curve ###
		# Curve Data
		c = self.convertToBezier()
		o = Blender.Object.New("Curve")
		# place it at the insertion cursor
		cursor_pos = Blender.Window.GetCursorPos()
		o.LocX = cursor_pos[0]
		o.LocY = cursor_pos[1]
		o.LocZ = cursor_pos[2]
		o.setName("Paint_Curve")

		o.link(c)
		sc.link(o)

	def convertToBezier(self):
		c = Blender.Curve.New("Paint_Curve")
		i = 0
		for stroke in self.strokes:
			stroke.convertToNurb(c, i)
			i += 1
		return(c)

class Canvas:
	def __init__(self):
		self.curve = Curve2D()
		self.cw = 0.0
		self.x = 0.0
		self.y = 0.0
		self.xmin = 0.0
		self.ymin = 0.0
		self.xmax = 0.0
		self.ymax = 0.0
		self.height = 0.0
		self.width = 0.0
	def draw_canvas(self):
		glClearColor(0.0, 0.0, 0.0, 0.0)
		glClear(GL_COLOR_BUFFER_BIT)
		# box dimensions designated for drawing
		box_size=Buffer(GL_FLOAT, 4)
    		glGetFloatv(GL_SCISSOR_BOX, box_size)
		self.x = box_size[0]
		self.y = box_size[1]
		width = box_size[2]
		height = box_size[3]
		# center white drawing area
		self.cw = min(width, height) / 2.0 - 20
		c = [ width / 2.0, height / 2.0]
		glColor3f(1.0, 1.0, 1.0)
		self.xmin = c[0] - self.cw
		self.ymin = c[1] - self.cw
		self.xmax = c[0] + self.cw
		self.ymax = c[1] + self.cw
		self.height = self.ymax - self.ymin
		self.width = self.xmax - self.xmin
		glRectf(self.xmin, self.ymin, self.xmax, self.ymax)
	def draw_contents(self):
		self.curve.draw()
	def paint(self, xi, yi):
		xi -= self.x
		yi -= self.y
		xi -= self.xmin
		yi -= self.ymin
		# fit mouse coordinates into the range [0.0 ... 1.0]
		x = xi / self.cw
		y = yi / self.cw
		self.curve.paint_stroke(x, y)
	def mouseIsOnCanvas(self, xi, yi):
		xi -= self.x
		yi -= self.y
		# check if mouse is on canvas
		if ( (xi > self.xmin) and (xi < self.xmax) and (yi > self.ymin) and (yi < self.ymax) ):
			return(Blender.TRUE)
		return(Blender.FALSE)
	def normCoordsToAreaCoords(self, x, y):
		xi = x * self.cw
		yi = y * self.cw
		xi = xi + self.xmin
		yi = yi + self.ymin
		return(xi, yi)


canvas = Canvas()


def finish_stroke():
	global stroke_flag
	canvas.curve.finish_stroke()	# The mouse button can be released outside the canvas as well
	stroke_flag = END_STROKE

def add_stroke():
	global stroke_flag
	stroke_flag = NEW_STROKE

def pencil_curve():
	global stroke_flag
	mouse_pos = Blender.Window.GetMouseCoords()
	if (canvas.mouseIsOnCanvas(mouse_pos[0], mouse_pos[1]) == Blender.TRUE):
		if (stroke_flag == DO_STROKE):
			canvas.paint(mouse_pos[0], mouse_pos[1])
		elif (stroke_flag == NEW_STROKE):
			canvas.curve.add_stroke()
			stroke_flag = DO_STROKE
		elif (stroke_flag == END_STROKE):
			stroke_flag = NO_STROKE
		elif (stroke_flag == NO_STROKE):
			pass
		else:
			print "bug in pencil_curve ???"


def draw():
	canvas.draw_canvas()

	canvas.curve.draw()

def event(evt, val):
	if (evt == QKEY and not val):
		Exit()
	elif (evt == XKEY and not val):
		canvas.curve.delete_strokes()
		Redraw()
	elif (evt == MKEY and not val):
		canvas.curve.make_real()
		Blender.Window.QRedrawAll()
	elif (evt == LEFTMOUSE):
		if (val == 0):
			finish_stroke()
		else:
			add_stroke()
	pencil_curve()

def bevent(evt):
	if (evt == 1):
		Exit()

# Bail out if the script is running as a script link
if (Blender.bylink):
	pass
else:
	Register(draw, event, bevent)

