/* BKE_lsystem.h */
#ifndef BKE_LSYSTEM_H
#define BKE_LSYSTEM_H

#define LSYS_MAX_LEN	256	/* ugly: defined in DNA_lsystem_types.h, too */
/* TODO:
	- pruning silhouettes
	- stochastic lsystems
	- tropism
*/
struct Lsystem;
struct LRule;
struct TurtleState;
struct LProfile;
struct DispListMesh;
struct ListBase;

typedef struct LChar {
	char c;
	char pad[3];
	short gen;		/* Which iteration has the LChar been created in ?*/
	short lifetime;		/* copy from LRule->lifetime */
	short flag;
	float param[4];
	ID *idp;			/* for implants */
} LChar;

/* lchar->flag */
#define LSYS_USEDEF	1	/* use the current default angle or length */

typedef struct LRule {
	struct LRule *next, *prev;
	char c[4];
	short flag;
	short len;	/* length of compiled replacement string */
	int lifetime;
	char repl[LSYS_MAX_LEN];
	LChar lrepl[LSYS_MAX_LEN];		/* compiled replacement string */
} LRule;

typedef enum {
	eGeneratorType_Skeleton,
	eGeneratorType_Mesh,
	eGeneratorType_Curve,
	eGeneratorType_Nurb,
	NUM_GENERATOR_TYPES,
} eGeneratorType;

typedef enum {
	/*
		If _eGeneratorFlag_Persistent_ is set, gen->freeData() is not called
		at the end of _execute_generator_, and a pointer to the generator
		is saved within the lsystem, allowing the generator data to be accessed
		later. The generator has to provide a valid _draw()_ function.
		Generator data is freed calling gen->freeData() when the lsystem itself is freed.
	*/
	eGeneratorFlag_Persistent = (1<<0),
	/*
		In loops that call initData() successively _eGeneratorFlag_Initilized_ tells
		if initData() has been called before.
	*/
	eGeneratorFlag_Initialized = (1<<1),
} eGeneratorFlag;

typedef struct LGenerator {
	struct LGenerator *next, *prev;
	eGeneratorType type;
	eGeneratorFlag flag;
	/* _initData_ has to set the initial turtle state as well */
	void (*initData)(struct LGenerator *lg, struct TurtleState *ts);
	void (*freeData)(struct LGenerator *lg);
	void (*cmd_popTurtleState)(struct LGenerator *lg, struct TurtleState *ts);
	/* cmd_popTurtleState_pre() is called before the old state is popped */
	void (*cmd_popTurtleState_pre)(struct LGenerator *lg, struct TurtleState *ts);
	void (*cmd_pushTurtleState)(struct LGenerator *lg, struct TurtleState *ts);
	void (*cmd_F)(struct LGenerator *lg, struct TurtleState *ts);
	void (*cmd_f)(struct LGenerator *lg, struct TurtleState *ts);
	void (*cmd_I)(struct LGenerator *lg, struct TurtleState *ts, struct Object *ob);
	void (*cmd_curlyOpen)(struct LGenerator *lg, struct TurtleState *ts);
	void (*cmd_curlyClose)(struct LGenerator *lg, struct TurtleState *ts);
	/* optional */
	int (*draw)(struct LGenerator *lg, struct Lsystem *ls);
	void (*render)(struct LGenerator *lg);
} LGenerator;

/* Skeleton */
typedef struct LVertex {
	float pos[3];
	MCol mcol;
	short mati;
	short flag;
} LVertex;

/* LVertex->flag */
#define LSYS_DRAW	1	/* draw connecting geometry from the last vertex to the next. */
#define LSYS_POLY	2
#define LSYS_SPLIT	4


typedef struct LGeneratorSkeleton {
	LGenerator gen;

	LVertex *lv;
	LVertex *lvert;
	int totlvert;
	int alloclvert;
	short lvflag;

	ListBase implants;
} LGeneratorSkeleton;

/* Mesh */
typedef struct LPolyVert {
	float co[2];
} LPolyVert;

typedef struct LPolygon {
	int totvert;
	float projection[4][4]; 	/* polygons have to lay on one plane */
	LPolyVert *verts;
} LPolygon;

typedef struct LGeneratorMesh {
	LGenerator gen;

	struct DispListMesh *dlm;
	struct Mesh *contour;
	struct Mesh *flower;
	short state_flag;	/* LSYS_POP and LSYS_POLY */
	int mati;
	/* to speed up linking of segments */
	struct MEdge *sel_edges;
	int num_sel_edges;		/* How many edges are selected? */
	int prev_seg;
	/* distribution map */
	short dmap_flag;	/* LSYS_DISTRIBUTION_MAP, LSYS_ORIENT_TO_NORMAL */
	float dmap_m[4][4];
} LGeneratorMesh;


struct Lsystem *add_lsystem(void);
struct Lsystem *copy_lsystem(struct Lsystem *ls);
void make_local_lsystem(struct Lsystem *ls);

void add_rule_v(void *ls_v, void *c_v);
void delete_rule(void *ls_v, void *rl_v);
void check_empty_axiom(void *ls_v, void *dummy);
void check_unique_rule_name(void *ls_v, void *c_v);

void boundbox_lsystem(struct Lsystem *ls, float *loc, float *size);

void lsystem_update(struct Object *ob, struct Lsystem *ls);

void add_persistent_generator(struct Lsystem *ls, LGenerator *gen);
void rem_persistent_generator(struct Lsystem *ls, eGeneratorType gtype);
LGenerator *fetch_persistent_generator(struct Lsystem *ls, eGeneratorType gtype);
void initGenerator_callbacks(LGenerator *lg, eGeneratorType gtype);

void execute_generator(struct Lsystem *ls, LGenerator *lgen);

struct DispListMesh *distribution_map_loop(struct Lsystem *ls, struct LGenerator *gen, struct DispListMesh *dmap);

void free_lsystem(struct Lsystem *ls);

void load_lsystem_from_file(char *name);
void save_lsystem_to_file(char *name);
void initLsystem_afterLoad(struct Lsystem *ls);

void lrules_to_saverules(struct ListBase *lrules, struct ListBase *saverules);
void saverules_to_lrules(struct ListBase *lrules, struct ListBase *saverules);

#endif
