/* lsystem.c */
#include <ctype.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "DNA_ID.h"
#include "DNA_lsystem_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_curve_types.h"
#include "DNA_object_types.h"
#include "DNA_scene_types.h"
#include "DNA_text_types.h"

#include "MEM_guardedalloc.h"

#include "BLI_blenlib.h"
#include "BLI_rand.h"
#include "MTC_matrixops.h"

#include "BIF_gl.h"
#include "BIF_glutil.h"
#include "BIF_screen.h"		/* for qtest() */

#include "BKE_curve.h"
#include "BKE_depsgraph.h"
#include "BKE_displist.h"
#include "BKE_global.h"
#include "BKE_ipo.h"
#include "BKE_library.h"
#include "BKE_lsystem.h"
#include "BKE_main.h"
#include "BKE_mesh.h"
#include "BKE_object.h"
#include "BKE_utildefines.h"

#include "blendef.h"
#include "butspace.h"
#include "mydevice.h"


#define LSTEST(nr) \
	fprintf(stderr, "Test %d\n", nr);	\
	fflush(stderr)

typedef enum EErrCode {
	LS_ERROR,
	LS_WARNING,
	LS_BUG,
	LS_DEBUG,
} EErrCode;

static void ls_error(EErrCode err, char *msg, ...)
{
	char	tmp[1024];
	const char	errtype[4][30] = {
			"Lsystem Error: ",
			"Lsystem Warning: ",
			"Lsystem Internal Bug: ",
			""
			};
	va_list	args;

	va_start(args,msg);
	vsprintf(tmp,msg,args);
	fprintf(stderr, "%s%s", errtype[err], tmp);
	fflush(stderr);
}

static void print_matrix(float m[][4])
{
	fprintf(stderr, "[\n%f %f %f %f\n%f %f %f %f\n%f %f %f %f\n%f %f %f %f\n",
			m[0][0], m[0][1], m[0][2], m[0][3],
			m[1][0], m[1][1], m[1][2], m[1][3],
			m[2][0], m[2][1], m[2][2], m[2][3],
			m[3][0], m[3][1], m[3][2], m[3][3]
			);
	fprintf(stderr, "]\n#");
}

/* deb_flag  */
#define	TRANS_MAT	1
#define PROD_STR	2
#define TURTLE_STACK	4
#define IMPLANTS	8
#define RULES		16
#define ALL		TRANS_MAT + PROD_STR + TURTLE_STACK + IMPLANTS

static void debug_lsystem(Lsystem *ls, short deb_flag)
{
	LChar *lc;
	LVertex *lv;
	TurtleState *ts;
	LRule *rl;
	int i;
	if (deb_flag & TRANS_MAT) {
		fprintf(stderr, "\nLsystem: \n");
		fprintf(stderr, "Transformation Matrix:\n");
		print_matrix(ls->curts.m);
	}
	if (deb_flag & PROD_STR) {
		fprintf(stderr, "Production String:\n");
		lc= ls->prod;
		i= ls->plen;
		while(i--) {
			if (lc->c == 'F')
				fprintf(stderr, "%c(%d)", lc->c, lc->gen);
			else
				fprintf(stderr, "%c", lc->c);
			lc++;
		}
		fprintf(stderr, "\n");
	}
	if (deb_flag & TURTLE_STACK) {
		fprintf(stderr, "Turtle Stack:\n");
		ts= (TurtleState *) ls->tstates.first;
		i= 0;
		while(ts) {
			fprintf(stderr, "%d:\n", i);
			print_matrix(ts->m);
			fprintf(stderr, "Color (r, g, b, a): %d, %d, %d, %d\n", ts->mcol.r, ts->mcol.g, ts->mcol.b, ts->mcol.a);

			ts= (TurtleState *) ts->next;
			i++;
		}
	}
	if (deb_flag & IMPLANTS) {

	}
	if (deb_flag & RULES) {
		fprintf(stderr, "Rules: \n");
		rl= ls->rules.first;
		while(rl) {
			fprintf(stderr, "%c --> %s\n", rl->c[0], rl->repl);
			rl= rl->next;
		}
	}
	fflush(stderr);
}

#define REGISTER_SKELETON_GENERATOR(ls, lgen)	\
	lgen= (LGenerator *) MEM_callocN(sizeof(LGeneratorSkeleton), "lgen register_skel");	\
	initGenerator_callbacks(lgen, eGeneratorType_Skeleton);		\
	add_persistent_generator(ls, lgen)

Lsystem *add_lsystem(void)
{
	Lsystem *ls;
	LGenerator *lgen;

	ls= alloc_libblock(&G.main->lsystem, ID_LS, "Lsystem");
	/* ls is zeroed by calloc in alloc_libblock */
	ls->filename[0]= '\0';

	ls->bb= unit_boundbox();
	/* defaults */
	strcpy(ls->axiom, "F+F+F+F+F");
	ls->iter= 3;
	ls->defangle= 30;
	ls->angle_step= 0.1;
	ls->deflen= 1.0;
	ls->len_step= 0.1;
	ls->fperiter= 10;

	ls->numrules= 0;

	ls->growtype= LSYS_LINEAR;
	ls->growsta= 0.0;
	ls->growend= 1.0;

	ls->lod_minsize= 1.0;

	ls->drawflag |= LSYS_DRAWSKEL + LSYS_DRAWGEOM + LSYS_DRAWIMPL;
	ls->flag |= LSYS_GROWLEN;
	/* LSYS_COMPILED is not set */
	REGISTER_SKELETON_GENERATOR(ls, lgen);

	return(ls);
}


Lsystem *copy_lsystem(Lsystem *ls) {
	Lsystem *lsn;
	LRule *rl, *rln;
	LGenerator *lgen;

	lsn= (Lsystem *) copy_libblock(ls);
	lsn->prod= lsn->prodn= lsn->compaxiom= NULL;
	/* up to this point no bug */
	lsn->rules.first= lsn->rules.last= NULL;
	lsn->generators.first= lsn->generators.last= NULL;
	REGISTER_SKELETON_GENERATOR(lsn, lgen);

	rl= (LRule *) ls->rules.first;
	while(rl) {
		rln= (LRule *) MEM_dupallocN(rl);
		BLI_addtail(&lsn->rules, rln);
		rl= rl->next;
	}

	lsn->flag &= ~LSYS_COMPILED;
	return(lsn);
}

void make_local_lsystem(Lsystem *ls)
{
	Object *ob;
	Lsystem *lsn;
	int local=0, lib=0;

	/* - only lib users: do nothing
	    * - only local users: set flag
	    * - mixed: make copy
	    */

	if(ls->id.lib==0) return;
	if(ls->id.us==1) {
		ls->id.lib= 0;
		ls->id.flag= LIB_LOCAL;
		new_id(0, (ID *)ls, 0);
		return;
	}

	ob= G.main->object.first;
	while(ob) {
		if(ob->data==ls) {
			if(ob->id.lib) lib= 1;
			else local= 1;
		}
		ob= ob->id.next;
	}

	if(local && lib==0) {
		ls->id.lib= 0;
		ls->id.flag= LIB_LOCAL;
		new_id(0, (ID *)ls, 0);
	}
	else if(local && lib) {
		lsn= copy_lsystem(ls);
		lsn->id.us= 0;

		ob= G.main->object.first;
		while(ob) {
			if(ob->data==ls) {

				if(ob->id.lib==0) {
					ob->data= lsn;
					lsn->id.us++;
					ls->id.us--;
				}
			}
			ob= ob->id.next;
		}
	}
}

void initLsystem_afterLoad(Lsystem *ls) {
	LGenerator *lgen;

	/* undefine temporal data so that it is recalculated */
	ls->prod= ls->prodn= ls->compaxiom= NULL;
	ls->flag &= ~LSYS_COMPILED;
	ls->generators.first= ls->generators.last= NULL;
	REGISTER_SKELETON_GENERATOR(ls, lgen);
}

static LRule *add_rule(Lsystem *ls, char *c)
{
	LRule *rl;
	rl= (LRule *) MEM_callocN(sizeof(LRule), "lsys add_rule");
	BLI_addtail(&ls->rules, rl);
	ls->numrules++;
	rl->c[0]= *c;
	rl->c[1]= '\0';
	rl->flag = 0;
	rl->lifetime= -1;
	rl->repl[0]= '\0';

	check_unique_rule_name(ls, c);
	return(rl);
}

/* for uiButSetFunc */
void add_rule_v(void *ls_v, void *c_v)
{
	Lsystem *ls= (Lsystem *) ls_v;
	char *c= (char *) c_v;

	add_rule(ls, c);
}

void delete_rule(void *ls_v, void *rl_v)
{
	Lsystem *ls= (Lsystem *) ls_v;
	LRule *rl= (LRule *) rl_v;
	BLI_freelinkN(&ls->rules, rl);
	ls->numrules--;
}

static void push_turtle_state(Lsystem *ls)
{
	TurtleState *ts;

	ts= (TurtleState *) MEM_mallocN(sizeof(TurtleState), "lsys push_tstate");
	memcpy(ts, &ls->curts, sizeof(TurtleState) );
	BLI_addtail(&ls->tstates, ts);

}

static void pop_turtle_state(Lsystem *ls)
{
	if (!ls->tstates.last) {
		ls_error(LS_ERROR, ">>] Cannot pop empty turtle stack!\n");
		return;
	}
	memcpy(&ls->curts, ls->tstates.last, sizeof(TurtleState) );
	BLI_freelinkN(&ls->tstates, ls->tstates.last);
}

void check_empty_axiom(void *ls_v, void *dummy)
{
	Lsystem *ls= (Lsystem *) ls_v;
	if (ls->axiom[0] == '\0') {
		strcpy(ls->axiom, "F");
	}
}

void check_unique_rule_name(void *ls_v, void *c_v)
{
	LRule *rl;
	LRule *drl;

	Lsystem *ls= (Lsystem *) ls_v;
	char *c= (char *) c_v;

	int found= 0;
	int isal;

	rl= (LRule *) ls->rules.first;
	while(rl) {
		if(rl->c[0] == *c) {
			found++;
			drl= rl;
		}
		rl= rl->next;
	}
	if(found > 1) {
		drl->c[0]= '\0';
	} else if (found== 1) {
		/* check if the rule name is alphanumeric */
		isal= isalpha((int) *c);
		if(!isal) {
			ls_error(LS_ERROR, "Rule letter is not a character of the alphabet!\n");
			drl->c[0]= '\0';
		} else {
			drl->c[0]= *c;
			drl->c[1]= '\0';
		}
	} else {
		ls_error(LS_BUG, "check_unique_rule_name ???\n");
	}
}

#define GET_MOD_BRACKETS(c, modi, delim1, delim2); 	\
	modi[0]= '\0';					\
	if (*(c+1) == delim1 ) {				\
		i= 0;				\
		st = c+1;				\
		while(*c != '\0') {		\
			if(*c == delim2 ) break;	\
			i++;			\
			c++;			\
		}				\
		if(*c == '\0') {		\
			ls_error(LS_ERROR, "Brackets not closed before end of production string\n");	\
		} else {			\
			memcpy(modi, st, i);	\
			modi[i]= '\0';	\
		}				\
	}

#define MOD_F1(c, modi, l);					\
	GET_MOD_BRACKETS(c, modi, '(', ')');				\
	if (modi[0] == '\0')	{					\
		l->flag |= LSYS_USEDEF;					\
	} else {							\
		scanret= sscanf(modi, "(%f)", &l->param[0]);		\
		if (scanret != 1) {					\
			ls_error(LS_ERROR, "One float expected e.g. (10.0)\n");	\
			l->flag |= LSYS_USEDEF;				\
		}							\
		l->flag &= ~LSYS_USEDEF;					\
	}        \

#define MOD_F3(c, modi, l);					\
	GET_MOD_BRACKETS(c, modi, '(', ')');				\
	if (modi[0] == '\0')	{					\
		l->flag |= LSYS_USEDEF;					\
	} else {							\
		scanret= sscanf(modi, "(%f, %f, %f)", &l->param[0], &l->param[1], &l->param[2]);		\
		if (scanret != 3) {					\
			ls_error(LS_ERROR, "Three numbers expected e.g. (0.1, 0.8, -12.3)\n");	\
			l->flag |= LSYS_USEDEF;				\
		}							\
		l->flag &= ~LSYS_USEDEF;					\
	}

#define MOD_F4(c, modi, l);					\
	GET_MOD_BRACKETS(c, modi, '(', ')');				\
	if (modi[0] == '\0')	{					\
		l->flag |= LSYS_USEDEF;					\
	} else {							\
		scanret= sscanf(modi, "(%f, %f, %f, %f)", &l->param[0], &l->param[1], &l->param[2], &l->param[3]);		\
		if (scanret != 4) {					\
			ls_error(LS_ERROR, "Four numbers [0.0 ... 1.0] expected e.g. (0.1, 0.8, 1.0, 0.23)\n");	\
			l->flag |= LSYS_USEDEF;				\
		}							\
		l->flag &= ~LSYS_USEDEF;					\
	}

#define MOD_OBJECT(c, modi, l);					\
	GET_MOD_BRACKETS(c, modi, '(', ')');				\
	if (modi[0] == '\0')	{					\
		ls_error(LS_ERROR, "Object name expected e.g. ( I(Mesh1) )\n");	\
		l->idp= NULL;						\
	} else {							\
		int string_len;						\
									\
		string_len= strlen(modi) - 2;				\
		strncpy(name, modi+1, string_len);  			\
		name[string_len]= '\0';					\
		test_obpoin_but(name, &l->idp);				\
		if(!l->idp) {						\
			ls_error(LS_ERROR, "Implant object %s does not exist\n", name);	\
		}							\
	}								\

typedef struct Variable {
	char name[16];
	float value;
} Variable;

static Variable variables[]= {
	{ "PI",		3.14 },
	{ "SQRT2",	1.76 },
	{ "",	0.0	},
};


static int getVariable_F1(char *modi, LChar *l)
{
	int scanret;
	int i= 0;
	Variable *v;

	v= &variables[i];
	while(v->name[0] != '\0') {
		v= &variables[i++];
		if (strcmp(v->name, modi)==0) {
			l->param[0]= v->value;
			break;
		}
	}
}

static int compile_string(char *str, LChar *lstr)
{
	char *c;
	LChar *l;
	char modi[256];
	char name[256];	/* needed in macro */
	int len= 0;
	int i;		/* needed in macro */
	char *st;	/* needed in macro */
	int scanret;	/* needed in macro */

	c= str;
	l= lstr;
	while(*c != '\0') {
		l->flag = 0;
		switch(*c) {
		/* Movement commands */
		case 'F':
			l->c= 'F';
			MOD_F1(c, modi, l);
			break;
		case 'f':
			l->c= 'f';
			MOD_F1(c, modi, l);
			break;
		/* Special commands */
		case 'I':
			l->c= 'I';
			MOD_OBJECT(c, modi, l);
			break;
		case 'c':
			l->c= 'c';
			MOD_F4(c, modi, l);
			break;
		case 'm':
			l->c= 'm';
			MOD_F1(c, modi, l);
			break;
		/* Turtle orientation commands */
		case '+':
			l->c= '+';
			MOD_F1(c, modi, l);
			break;
		case '-':
			l->c= '-';
			MOD_F1(c, modi, l);
			break;
		case '<':
			l->c= '<';
			MOD_F1(c, modi, l);
			break;
		case '>':
			l->c= '>';
			MOD_F1(c, modi, l);
			break;
		case '&':
			l->c= '&';
			MOD_F1(c, modi, l);
			break;
		case '^':
			l->c= '^';
			MOD_F1(c, modi, l);
			break;
		/* Special orientation commands */
		case '|':
			l->c= '+';
			l->param[0]= 180.0;
			break;
		case '%':
			l->c= '<';
			l->param[0]= 180.0;
			break;
		case '~':			/* not yet implemented */
			l->c= '~';
			MOD_F1(c, modi, l);
			break;
		case 't':			/* not yet implemented */
			l->c= 't';
			MOD_F1(c, modi, l);
			break;
		/* increment/decrement length */
		case '"':
			l->c= '"';
			MOD_F1(c, modi, l);
			break;
		case '\'':
			l->c= '\'';
			MOD_F1(c, modi, l);
			break;
		/* increment/decrement angle */
		case ';':
			l->c= ';';
			MOD_F1(c, modi, l);
			break;
		case ':':
			l->c= ':';
			MOD_F1(c, modi, l);
			break;
		/* direct matrix manipulation: scale */
		case 's':
			l->c= 's';
			MOD_F3(c, modi, l);
			break;
		/* ignore */
		case ' ':
			c++;
			continue;
		default:
			l->c= *c;
			break;
		};
		l++;
		len++;
		/* BUG: check if *c == '\0' */
		c++;
	}

	return(len);
}

void compile_lsystem(Lsystem *ls)
{
	LRule *rl;
	LChar *lc;
	int a;

	/* compile rules */
	rl= (LRule *) ls->rules.first;
	while(rl) {
		rl->len= compile_string(rl->repl, rl->lrepl);
		/* set flag if set in rule*/
		a= rl->len;
		lc= rl->lrepl;
		while(a--) {
			lc->flag |= rl->flag;
			lc->lifetime= rl->lifetime;
			lc++;
		}
		rl= rl->next;
	}
	/* compile axiom */
	ls->complen= 0;
	if (ls->compaxiom) {
		MEM_freeN(ls->compaxiom);
	}
	ls->compaxiom= MEM_mallocN(sizeof(LChar) * LSYS_MAX_LEN, "lsys ls->compaxiom");
	ls->complen= compile_string(ls->axiom, ls->compaxiom);

	a= ls->complen;
	lc= ls->compaxiom;
	while(a--) {
		lc->gen= 0;
		lc->lifetime= -1;
		lc++;
	}

	ls->flag |= LSYS_COMPILED;
}

static void set_lifetime(LChar *lc, LChar *lrepl, int len)
{
	if (lrepl->lifetime < lc->lifetime) {
		while(len--) {
			lrepl->lifetime= lc->lifetime;
		}
	}
}

static void set_generation(LChar *lc, LChar *lrepl, int rlen, int iter, int lastplen, int plen)
{
	LChar *lptr;
	int a;

	lptr= lrepl;
	a= rlen;
	while(a--) {
		lptr->gen= lc->gen + 1;
		lptr++;
	}

	if (plen < lastplen) {
		if (lrepl->c == lc->c) {
			lrepl->gen= lc->gen;
		}
	}
}

static void create_prod_string(Lsystem *ls)
{
	LRule *rl;
	LChar *l;
	char c;
	LChar origc;
	LChar *lrepl;
	int i, plen, len, rlen, iter;
	int lastplen;
	unsigned short event;

	len= 0;

	iter= ls->iter;
/* This seems to be rubbish
        if (ls->flag & LSYS_GROWING) {
		//	If generations do not grow stippled, we can skip unecessary
		//   	production string iterations
		if ( (ls->growsta >= 0.0) && (ls->growend <= 1.0) )
			iter= floor(ls->gentime);
	}
*/
	if (ls->prod) {
		MEM_freeN(ls->prod);
	}
	ls->prod= MEM_mallocN(sizeof(LChar) * LSYS_MAX_LEN, "lsys ls->prod");
	memcpy(ls->prod, ls->compaxiom, ls->complen * sizeof(LChar) );
	ls->plen= ls->complen;
	for(i= 0; i<iter; i++) {
		l= ls->prod;

		plen= ls->plen;

		lastplen= plen;

		ls->plen= 0;
		/* check if user has hit ESCAPE */
		event= qtest();
		if (event==ESCKEY) {
			ls_error(LS_WARNING, "User has hit Escape.\n");
			return;
		}

		while(plen--) {
			c= l->c;
			memcpy(&origc, l, sizeof(LChar) ); rlen= 1;

			lrepl= &origc;

			rl= (LRule *) ls->rules.first;
			while(rl) {
				if(rl->c[0]==c) {
					if ( (l->gen < l->lifetime) || (l->lifetime==-1) ) {
						lrepl= rl->lrepl;
						rlen= rl->len;
						break;
					}
				}
				rl= rl->next;
			}
			if(len <= ls->plen + rlen) {
				LChar *newprod;
				len += (1024 > rlen) ? 1024 : rlen;
				/* realloc(len) */
				newprod= (LChar *) MEM_mallocN(len*sizeof(LChar), "grow prodstr");
				if (ls->prodn) {
					memcpy(newprod, ls->prodn, ls->plen * sizeof(LChar) );
					MEM_freeN(ls->prodn);
				}
				/* */
				ls->prodn= newprod;
			}
			memcpy(ls->prodn + ls->plen, lrepl, sizeof(LChar) * rlen);
			set_generation(l, ls->prodn + ls->plen, rlen, i, lastplen, ls->plen);
			set_lifetime(l, ls->prodn + ls->plen, rlen);
			ls->plen += rlen;

			l++;
		}
		MEM_freeN(ls->prod);
		ls->prod= ls->prodn;
		ls->prodn= NULL;
		len= 0;
	}
	//debug_lsystem(ls, PROD_STR);
}

static void generation_time(Object *ob, Lsystem *ls)
{
	float ctime;

	if (ls->flag & LSYS_GROWING) {
		ctime= bsystem_time(ob, 0, (float)G.scene->r.cfra, 0);
		ls->gentime= ctime/ls->fperiter;
		ls->gentime= MIN2(ls->gentime, ls->iter + 1);
	}
}


void lsystem_update(Object *ob, Lsystem *ls)
{
	if ( !(ls->flag & LSYS_COMPILED) )
		compile_lsystem(ls);
	generation_time(ob, ls);
	create_prod_string(ls);

	if (ls->bb) MEM_freeN(ls->bb);
	ls->bb= NULL;
}

/* Transformation Routines */
static void Mat4Rotate(float m[][4], float angle, float axis[3])
{
	float s, c, t, x_2, y_2, z_2;

	s = sin(angle);
	c = cos(angle);
	t = 1 - c;

#define x axis[0]
#define y axis[1]
#define z axis[2]
	x_2 = x*x;
	y_2 = y*y;
	z_2 = z*z;

	float mtransf[4][4] = {
		{ x_2 + c * (1-x_2),	x*y*t - z*s,		x*z*t + y*s,		0 },
		{ x*y*t + z*s,		y_2 + c * (1-y_2),	y*z*t - x*s,		0 },
		{ x*z*t - y*s,		y*z*t + x*s,		z_2 + c * (1-z_2),	0 },
		{ 0,			0,			0,			1 },
	};

	float mnew[4][4];

	//MTC_Mat4MulMat4(mnew, m, mtransf);
	MTC_Mat4MulMat4(mnew, mtransf, m);
	MTC_Mat4CpyMat4(m, mnew);
#undef x
#undef y
#undef z
}

static void Mat4Translate(float m[][4], float x, float y, float z)
{
	float mtransf[4][4] = {
		{ 1, 0, 0, 0 },
		{ 0, 1, 0, 0 },
		{ 0, 0, 1, 0 },
		{ x, y, z, 1 }
		};
	float mnew[4][4];

	MTC_Mat4MulMat4(mnew, mtransf, m);
	MTC_Mat4CpyMat4(m, mnew);
}

static void Mat4Scale(float m[][4], float sx, float sy, float sz)
{
	float mtransf[4][4] = {
		{ sx, 0, 0, 0 },
		{ 0, sy, 0, 0 },
		{ 0, 0, sz, 0 },
		{ 0, 0, 0, 1  }
		};
	float mnew[4][4];

	MTC_Mat4MulMat4(mnew, mtransf, m);
	MTC_Mat4CpyMat4(m, mnew);
}

#define RADIANS(degree) \
	(degree) * M_PI / 180


#define VECTOR3(v, x, y, z)	\
	v[0] = (x);		\
	v[1] = (y);		\
	v[2] = (z)


#define MCOLOR(mcol, rc, gc, bc, ac)	\
	mcol.r = (unsigned char) ((rc) * 255);	\
	mcol.g = (unsigned char) ((gc) * 255);	\
	mcol.b = (unsigned char) ((bc) * 255);	\
	mcol.a = (unsigned char) ((ac) * 255)


#define CHECK_USEDEF_F1(l, def);	\
	if (l->flag & LSYS_USEDEF) {			\
		l->param[0]= def;				\
	}

#define CHECK_USEDEF_F3(l, def0, def1, def2)		\
	if (l->flag & LSYS_USEDEF) {			\
		l->param[0]= def0;			\
		l->param[1]= def1;			\
		l->param[2]= def2;			\
	}

#define CHECK_USEDEF_F4(l, def0, def1, def2, def3)	\
	if (l->flag & LSYS_USEDEF) {			\
		l->param[0]= def0;			\
		l->param[1]= def1;			\
		l->param[2]= def2;			\
		l->param[3]= def3;			\
	}

#define GROWING_FRAC(l, gflag);	\
	param= l->param[0];			\
	if (ls->flag & LSYS_GROWING) {		\
		if (ls->flag & gflag) {		\
			param= l->param[0] * precalc_growfrac[l->gen];	\
		}				\
	}

#define GROWING_FRAC_F3(l, gflag); \
	if (ls->flag & LSYS_GROWING) {		\
		if (ls->flag & gflag) {		\
			l->param[0] *= precalc_growfrac[l->gen];	\
			l->param[1] *= precalc_growfrac[l->gen];	\
			l->param[2] *= precalc_growfrac[l->gen];	\
		}				\
	}

static float calc_growfrac(Lsystem *ls, float growtime)
{
	float x, y;
	const float G= 1.0;
	const float y_0= 0.00001;
	const float k= 1.0;

	x= growtime;
	if (ls->growsta >= ls->growend) {
		ls->growsta= 0.0;
		ls->growend= 1.0;
		ls_error(LS_WARNING, "gsta has to be < gend!\n");
	}
	CLAMP(x, ls->growsta, ls->growend);
	x= (x - ls->growsta) * 1 / (ls->growend - ls->growsta);

	switch(ls->growtype) {
	case LSYS_LINEAR:
		y= x;
		return( y );
	case LSYS_RESTRICTED:
		y= -2*x*x + 4*x - 1;
		return( y );
	case LSYS_LOGISTIC:
		y= 1 / (1 + expf(-k*G*x)*(G/y_0 - 1) );
	default:
		ls_error(LS_BUG, "calc_growfrac ???\n");
		return( growtime );
	}

}

void boundbox_lsystem(Lsystem *ls, float *loc, float *size)
{
	LGeneratorSkeleton *gen;
	LVertex *lv;
	BoundBox *bb;
	float min[3], max[3];
	float mloc[3], msize[3];
	int a;

	ls_error(LS_DEBUG, "boundbox_lsystem\n");
	if(ls->bb==0) ls->bb= MEM_callocN(sizeof(BoundBox), "lsys boundbox");
	bb= ls->bb;

	INIT_MINMAX(min, max);

	if (!loc) loc= mloc;
	if (!size) size= msize;

	loc[0]= loc[1]= loc[2]= 0.0;
	size[0]= size[1]= size[2]= 0.0;

	gen= (LGeneratorSkeleton *) fetch_persistent_generator(ls, eGeneratorType_Skeleton);
	if (gen) {
		lv= gen->lvert;
		for(a=0; a<gen->totlvert; a++, lv++) {
			DO_MINMAX(lv->pos, min, max);
		}

		if(gen->totlvert) {
			loc[0]= (min[0]+max[0])/2.0f;
			loc[1]= (min[1]+max[1])/2.0f;
			loc[2]= (min[2]+max[2])/2.0f;

			size[0]= (max[0]-min[0])/2.0f;
			size[1]= (max[1]-min[1])/2.0f;
			size[2]= (max[2]-min[2])/2.0f;
		}
	}

	bb->vec[0][0]=bb->vec[1][0]=bb->vec[2][0]=bb->vec[3][0]= loc[0]-size[0];
	bb->vec[4][0]=bb->vec[5][0]=bb->vec[6][0]=bb->vec[7][0]= loc[0]+size[0];

	bb->vec[0][1]=bb->vec[1][1]=bb->vec[4][1]=bb->vec[5][1]= loc[1]-size[1];
	bb->vec[2][1]=bb->vec[3][1]=bb->vec[6][1]=bb->vec[7][1]= loc[1]+size[1];

	bb->vec[0][2]=bb->vec[3][2]=bb->vec[4][2]=bb->vec[7][2]= loc[2]-size[2];
	bb->vec[1][2]=bb->vec[2][2]=bb->vec[5][2]=bb->vec[6][2]= loc[2]+size[2];
}

/* Saving and Loading to a L-system description file */
static void parse_lsys_file(Lsystem *ls, FILE *fh, char *name)
{
	char keywd[64];
	char str[2048];
	char tmp[2048];
	int stat;
	int line= 1;
	LRule *rl;
	char c[2];	/* stores the rule character */
	char eq[2];	/* stores the '=' sign */
	const char parse_error_msg[]="Parse error in lsystem description in \"%s\" (line %d)!\n";

#define CHECK_STAT(exp);			\
	if (stat < (exp) ) {			\
		ls_error(LS_ERROR,		\
				(char *) parse_error_msg,	\
				name, line);	\
		line++;		/* because we skip this statement later */	\
		continue;			\
	}					\

	while(!feof(fh) ) {
		tmp[0]= '\0';		/* ???? otherwise the last line is parsed twice. */
		fgets(tmp, 2048, fh);
		if(tmp[0] == '#') {
			/* ignore comments */
		} else if ( (tmp[0] == '\n') || (tmp[0] == '\0') ) {
			/* ignore empty lines */
		} else {
			stat= sscanf(tmp, "%63s%2047s", keywd, str);
			CHECK_STAT(2);
			if(strcmp(keywd, "axiom") == 0) {
				strcpy(ls->axiom, str);
			} else if(strcmp(keywd, "rule") == 0) {
				stat= sscanf(tmp, "%63s%1s%1s%2047s", keywd, c, eq, str);
				CHECK_STAT(3);
				rl= add_rule(ls, c);
				strcpy(rl->repl, str);
			} else if(strcmp(keywd, "endrule") == 0) {
				stat= sscanf(tmp, "%63s%1s%1s%2047s", keywd, c, eq, str);
				CHECK_STAT(3);
				rl= add_rule(ls, c);
				strcpy(rl->repl, str);
				rl->lifetime = 1;
			} else if(strcmp(keywd, "iteration") == 0) {
				stat= sscanf(str, "%d", &ls->iter);
				CHECK_STAT(1);
			} else if(strcmp(keywd, "angle") == 0) {
				stat= sscanf(str, "%f", &ls->defangle);
				CHECK_STAT(1);
			} else if(strcmp(keywd, "length") == 0) {
				stat= sscanf(str, "%f", &ls->deflen);
				CHECK_STAT(1);
			} else if(strcmp(keywd, "color") == 0) {
				stat= sscanf(tmp, "%63s%f%f%f%f", keywd,
					&ls->defcolor[0], &ls->defcolor[1],
					&ls->defcolor[2], &ls->defcolor[3]
					);
				ls_error(LS_DEBUG, "stat %d\n", stat);
				CHECK_STAT(5);
			} else {
				ls_error(LS_ERROR,
					"Unknown keyword %s in lsystem description in \"%s\" (line %d)!\n",
					keywd, name, line);
			}
		}
		line++;
	}
}

void load_lsystem_from_file(char *name)
{
	Object *ob;
	Lsystem *ls;
	FILE *fh;
	LGenerator *lgen;

	ob= OBACT;
	if (ob==0) return;
	ls= (Lsystem *) ob->data;
	strncpy(ls->filename, name, MAX_FILE_LEN);
	/* reset this lsystem */
	free_lsystem(ls);
	/* persistent generators need to be added again, not nice */
	REGISTER_SKELETON_GENERATOR(ls, lgen);

	fh= fopen(name, "r");
	if(fh) {
		parse_lsys_file(ls, fh, name);
		fclose(fh);
	}
	ls->flag &= ~LSYS_COMPILED;
}

static void write_lsys_file(Lsystem *ls, FILE *fh, char *name)
{
	LRule *rl;

	fprintf(fh, "# %s\n", ls->id.name+2);
	fprintf(fh, "iteration %d\n", ls->iter);
	fprintf(fh, "angle %f\n", ls->defangle);
	fprintf(fh, "length %f\n", ls->deflen);
	fprintf(fh, "\n");
	fprintf(fh, "axiom %s\n", ls->axiom);

	rl= (LRule *) ls->rules.first;
	while(rl) {
		if (rl->lifetime != -1) {
			fprintf(fh, "endrule %c = %s\n", *rl->c, rl->repl);
		} else {
			fprintf(fh, "rule %c = %s\n", *rl->c, rl->repl);
		}
		rl= rl->next;
	}
}

void save_lsystem_to_file(char *name)
{
	Object *ob;
	Lsystem *ls;
	FILE *fh;

	ob= OBACT;
	if (ob==0) return;
	ls= (Lsystem *) ob->data;

	fh= fopen(name, "w");
	if(fh) {
		write_lsys_file(ls, fh, name);
		fclose(fh);
	}

}

/* lsystem_to_text */

/* lsystem_from_text */

/* Saving to a Blender file */

void lrules_to_saverules(ListBase *lrules, ListBase *saverules)
{
	LRule *rl;
	SaveRule *srl;

	saverules->first= saverules->last= NULL;
	rl= lrules->first;
	while(rl) {
		srl= (SaveRule *) MEM_mallocN(sizeof(SaveRule), "lrules2saverules");
		srl->c[0]= rl->c[0];
		srl->c[1]= '\0';
		srl->flag= rl->flag;
		srl->lifetime= (short) rl->lifetime;
		srl->len= strlen(rl->repl) + 1;
		srl->repl= MEM_mallocN(sizeof(char) * srl->len, "srl->repl");
		memcpy(srl->repl, rl->repl, srl->len);
		BLI_addtail(saverules, srl);

		rl= rl->next;
	}
}

void free_saverules(ListBase *saverules)
{
	SaveRule *srl;

	srl= (SaveRule *) saverules->first;
	while(srl) {
		MEM_freeN(srl->repl);
		MEM_freeN(srl);

		srl= (SaveRule *) srl->next;
	}
}

void saverules_to_lrules(ListBase *lrules, ListBase *saverules)
{
	LRule *rl;
	SaveRule *srl;

	lrules->first= lrules->last= NULL;
	srl= saverules->first;
	while(srl) {
		rl= (LRule *) MEM_mallocN(sizeof(LRule), "saverules2lrules");
		rl->c[0]= srl->c[0];
		rl->c[1]= '\0';
		rl->flag= srl->flag;
		rl->lifetime= (int) srl->lifetime;
		strncpy(rl->repl, srl->repl, LSYS_MAX_LEN);
		BLI_addtail(lrules, rl);

		srl= srl->next;
	}
}

/**********/

void free_lsystem(Lsystem *ls)
{
	LGenerator *gen;

	if(ls->prod) MEM_freeN(ls->prod);
	if(ls->compaxiom) MEM_freeN(ls->compaxiom);

	BLI_freelistN(&ls->rules);
	BLI_freelistN(&ls->tstates);

	gen= ls->generators.first;
	while(gen) {
		gen->freeData(gen);
		gen= gen->next;
	}
	BLI_freelistN(&ls->generators);

	if(ls->bb) MEM_freeN(ls->bb);
}


/********** Generators **********/
static TurtleState *default_turtle_state(TurtleState *ts)
{
	/* reset turtle state */
	MTC_Mat4One(ts->m);
	ts->sx= 1.0; ts->sy= 1.0; ts->sz= 1.0;

	ts->acc_F= 0.0;
	ts->acc_f= 0.0;
	ts->v.integer= 0;
	ts->mati= 1;
	return(ts);
}

/* in drawobject.c */
extern void draw_object(Base *base, int flag);

/* Skeleton */
#define GROW_LVERTS();			\
	if(lgen->totlvert == lgen->alloclvert) {	\
		lgen->alloclvert += 1024;		\
		LVertex *newlv= (LVertex *) MEM_mallocN(sizeof(LVertex) * lgen->alloclvert, "grow lverts");\
		if (lgen->lvert) {							\
			memcpy(newlv, lgen->lvert, sizeof(LVertex) * lgen->totlvert);	\
			MEM_freeN(lgen->lvert);						\
		}									\
		lgen->lvert= newlv;						\
	}									\

#define SET_LVERT(lvert, curts)		\
	/* set position */		\
	memcpy(&lv->pos, curts->m[3], 3*sizeof(float) );	\
	memcpy(&lv->mcol, &curts->mcol, sizeof(MCol) )

static void skeleton_initData(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	LVertex *lv;

	ts= default_turtle_state(ts);

	if ( (lgen->gen.flag & eGeneratorFlag_Initialized) && (lgen->lvert!=NULL)) {
		MEM_freeN(lgen->lvert);
	}
	lgen->lvert= NULL;
	lgen->alloclvert= 0;
	lgen->totlvert= 0;

	lgen->lvflag= 0;
	if( (lgen->gen.flag & eGeneratorFlag_Initialized) && (lgen->implants.first!=NULL)) {
		BLI_freelistN(&lgen->implants);
		lgen->implants.first= lgen->implants.last= NULL;
	}
	/* root vertex */
	GROW_LVERTS();
	lv= &lgen->lvert[lgen->totlvert++];
	VECTOR3(lv->pos, 0.0, 0.0, 0.0);
	lgen->lv= lv;

	lgen->gen.flag |= eGeneratorFlag_Initialized;
}

static void skeleton_freeData(LGenerator *gen)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;

	if (lgen->lvert) {
		MEM_freeN(lgen->lvert);
		lgen->lvert= NULL;
	}
	BLI_freelistN(&lgen->implants);
	lgen->implants.first= lgen->implants.last= NULL;

}

static void skeleton_cmd_popTurtleState(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	LVertex *lv;

	GROW_LVERTS();
	lv= &lgen->lvert[lgen->totlvert++];
	VECTOR3(lv->pos, ts->m[3][0], ts->m[3][1], ts->m[3][2]);
	lv->flag &= ~LSYS_DRAW;
	lv->flag |= LSYS_SPLIT;

	lgen->lv= lv;
}

static void skeleton_cmd_pushTurtleState(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;

	lgen->lv->flag |= LSYS_SPLIT;
}

static void skeleton_cmd_F(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	LVertex *lv;

	GROW_LVERTS();
	lv= &lgen->lvert[lgen->totlvert++];
	SET_LVERT(lv, ts);
	/* draw a connection from the previous vertex to this one */
	lv->flag |= LSYS_DRAW;
	lv->flag |= lgen->lvflag;

	lgen->lv= lv;
}

static void skeleton_cmd_f(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	LVertex *lv;

	GROW_LVERTS();
	lv= &lgen->lvert[lgen->totlvert++];
	/* set position only */
	memcpy(&lv->pos, ts->m[3], 3*sizeof(float) );
	lv->flag &= ~LSYS_DRAW;

	lgen->lv= lv;
}

static void skeleton_cmd_I(LGenerator *gen, TurtleState *ts, Object *ob)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	Object *newob;

	newob= MEM_mallocN(sizeof(Object), "newobj implant");
	memcpy(newob, ob, sizeof(Object));
	MTC_Mat4CpyMat4(newob->obmat, ts->m);
	/* TODO: Implants need to be scaled by ts->sx, ts->sy, ts->sz */
	BLI_addtail(&lgen->implants, newob);
}

static void skeleton_cmd_curlyOpen(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;

	lgen->lvflag= LSYS_POLY;

}

static void skeleton_cmd_curlyClose(LGenerator *gen, TurtleState *ts)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;

	lgen->lvflag= 0;
}

static int skeleton_draw(LGenerator *gen, Lsystem *ls)
{
	LGeneratorSkeleton *lgen= (LGeneratorSkeleton *) gen;
	int i;

	if (ls->drawflag & LSYS_DRAWSKEL) {
		glBegin(GL_LINES);
			for(i= 1; i<lgen->totlvert; i++) {
				glColor4f(	lgen->lvert[i].mcol.r/255.0f,
						lgen->lvert[i].mcol.g/255.0f,
						lgen->lvert[i].mcol.b/255.0f,
						lgen->lvert[i].mcol.a/255.0f
						);
				if (lgen->lvert[i].flag & LSYS_DRAW) {
					glVertex3f(	lgen->lvert[i-1].pos[0],
							lgen->lvert[i-1].pos[1],
							lgen->lvert[i-1].pos[2]
						);
				} else {
					glVertex3f(
						lgen->lvert[i].pos[0],
						lgen->lvert[i].pos[1],
						lgen->lvert[i].pos[2]
					);
				}
				glVertex3f(	lgen->lvert[i].pos[0],
						lgen->lvert[i].pos[1],
						lgen->lvert[i].pos[2]
					);
			}
		glEnd();
	}
	if(ls->drawflag & LSYS_DRAWIMPL) {
		Object *impl;
		Base tbase;

		impl= (Object *) lgen->implants.first;
		while(impl) {
			tbase.object= impl;
			draw_object(&tbase, 0);
			impl= (Object *) ((ID *) impl)->next;
		}
	}

	return(1);
}

/************************/
/* Mesh */
static MFace *cpadd_faces(DispListMesh *dlm, Mesh *sme)
{
	MFace *mface;
	int sta;
	int i;

	sta= dlm->totface;
	mface= (MFace *) MEM_callocN(sizeof(MFace) * (dlm->totface + sme->totface), "lsys new faces" );
	if (dlm->mface) {
		memcpy(mface, dlm->mface, sizeof(MFace)*dlm->totface);
		MEM_freeN(dlm->mface);
	}
	if (sme->mface) {
		memcpy(mface + dlm->totface, sme->mface, sizeof(MFace)*sme->totface);
	}
	dlm->mface= mface;
	dlm->totface += sme->totface;
	/* Fix indices */
	mface= &dlm->mface[sta];
	for(i= 0; i<sme->totface; i++, mface++) {
		mface->v1 += dlm->totvert;
		mface->v2 += dlm->totvert;
		mface->v3 += dlm->totvert;
		if (mface->v4) {
			mface->v4 += dlm->totvert;
		}
	}
	return(&dlm->mface[sta]);
}

static MEdge *cpadd_edges(DispListMesh *dlm, Mesh *sme)
{
	MEdge *medge;
	int sta;
	int i;

	sta= dlm->totedge;
	medge= (MEdge *) MEM_callocN(sizeof(MEdge) * (dlm->totedge + sme->totedge), "lsys new edges" );
	if (dlm->medge) {
		memcpy(medge, dlm->medge, sizeof(MEdge)*dlm->totedge);
		MEM_freeN(dlm->medge);
	}
	if (sme->medge)	{
		memcpy(medge + dlm->totedge, sme->medge, sizeof(MEdge)*sme->totedge);
	}
	dlm->medge= medge;
	dlm->totedge += sme->totedge;
	/* Fix indices */
	medge= &dlm->medge[sta];
	for(i= 0; i<sme->totedge; i++, medge++) {
		medge->v1 += dlm->totvert;
		medge->v2 += dlm->totvert;
	}
	return(&dlm->medge[sta]);
}

static void transform_vertices(DispListMesh *dlm, int sta, int nr, float m[][4])
{
	int i;
	MVert *mv;

	mv= &dlm->mvert[sta];
	for(i= 0; i< nr; i++, mv++) {
		MTC_Mat4MulVecfl(m, mv->co);
	}
}

static void merge_and_transform(DispListMesh *dlm, Mesh *add_mesh, TurtleState *ts, MEdge *medge, MFace *mface) {
	MVert *mvert;
	float ms[4][4];

	mvert= (MVert *) MEM_mallocN(sizeof(MVert) * (dlm->totvert + add_mesh->totvert), "merge_&_trans mvert");
	if (dlm->mvert) {
		memcpy(mvert, dlm->mvert, sizeof(MVert) * dlm->totvert);
		MEM_freeN(dlm->mvert);
	}
	dlm->mvert= mvert;
	if (add_mesh->mvert) {
		memcpy(dlm->mvert + dlm->totvert,
			add_mesh->mvert, sizeof(MVert)*add_mesh->totvert);
	}
	/* scale contour */
	memcpy(ms, ts->m, sizeof(float)*16);

	ms[0][0] *= ts->sx; ms[0][1] *= ts->sx; ms[0][2] *= ts->sx;
	ms[1][0] *= ts->sy; ms[1][1] *= ts->sy; ms[1][2] *= ts->sy;
	ms[2][0] *= ts->sz; ms[2][1] *= ts->sz; ms[2][2] *= ts->sz;
	transform_vertices(dlm, dlm->totvert, add_mesh->totvert, ms);
	/* merge faces and vertices */
	medge= cpadd_edges(dlm, add_mesh);
	mface= cpadd_faces(dlm, add_mesh);
}

static void add_segment(LGeneratorMesh *lgen, DispListMesh *dlm, Mesh *contour, TurtleState *ts) {
	MEdge *medge, *link_edges;
	MFace *mface;
	int i;
	int prev_seg, new_seg;

	if(!contour) {
		return;
	}
	merge_and_transform(dlm, contour, ts, medge, mface);
	/* Now we link the selected edges of the new segment and the previous one. */
	prev_seg= lgen->prev_seg;
	new_seg= dlm->totvert;
	dlm->totvert += contour->totvert;
	if (prev_seg >= 0) {
		/* Faces */
		mface= (MFace *) MEM_mallocN(sizeof(MFace) * (dlm->totface + lgen->num_sel_edges), "lsys linking faces");
		if (dlm->mface) {
			memcpy(mface, dlm->mface, sizeof(MFace)*dlm->totface);
			MEM_freeN(dlm->mface);
		}
		dlm->mface= mface;
		mface= dlm->mface + dlm->totface;
		medge= lgen->sel_edges;
		for(i= 0; i<lgen->num_sel_edges; i++, medge++, mface++) {
			mface->v1= medge->v1 + prev_seg;
			mface->v2= medge->v2 + prev_seg;
			mface->v4= medge->v1 + new_seg;
			mface->v3= medge->v2 + new_seg;
			mface->flag= medge->flag;
			mface->mat_nr= lgen->mati - 1;
		}
		dlm->totface += lgen->num_sel_edges;
		/* Edges */
		medge= (MEdge *) MEM_mallocN(sizeof(MEdge) * (dlm->totedge + lgen->num_sel_edges), "lsys linking edges");
		if (dlm->medge) {
			memcpy(medge, dlm->medge, sizeof(MEdge)*dlm->totedge);
			MEM_freeN(dlm->medge);
		}
		dlm->medge= medge;
		link_edges= dlm->medge + dlm->totedge;
		medge= lgen->sel_edges;
		for(i= 0; i<lgen->num_sel_edges; i++, medge++, link_edges++) {
			link_edges->v1= medge->v1 + prev_seg;
			link_edges->v2= medge->v1 + new_seg;
			link_edges->flag= medge->flag;
		}
		dlm->totedge += lgen->num_sel_edges;
	}
	lgen->prev_seg= new_seg;
}

static void make_cap(DispListMesh *dlm, float *tip_pos, int npnts)
{
	int len, i, j, k, p1, cap;
	MVert *mvert;
	MFace *mface;
	MEdge *medge;

	len= dlm->totvert + 1;
	mvert= (MVert *) MEM_mallocN(sizeof(MVert) * len, "make cap");
	if (dlm->mvert) {
		memcpy(mvert, dlm->mvert, sizeof(MVert) * dlm->totvert);
		MEM_freeN(dlm->mvert);
	}
	mvert[len-1].co[0]= tip_pos[0];
	mvert[len-1].co[1]= tip_pos[1];
	mvert[len-1].co[2]= tip_pos[2];

	//mface= new_faces(dlm, npnts);
	//medge= new_edges(dlm, npnts*2);

	cap= dlm->totvert;
	p1= dlm->totvert - npnts;
	for(i= p1, j= 0, k= 0; i < dlm->totvert-1; i++, j++, k++) {
		mface[j].v1= i;
		mface[j].v2= i+1;
		mface[j].v3= cap;
		mface[j].v4= i;

		medge[k].v1= i;
		medge[k].v2= i+1;
		medge[k].flag= ME_EDGEDRAW;
		k++;
		medge[k].v1= i;
		medge[k].v2= cap;
		medge[k].flag= ME_EDGEDRAW;
	}
	/* cyclic */
	mface[j].v1= i;
	mface[j].v2= p1;
	mface[j].v3= cap;
	mface[j].v4= i;

	medge[k].v1= i;
	medge[k].v2= p1;
	medge[k].flag= ME_EDGEDRAW;
	k++;
	medge[k].v1= i;
	medge[k].v2= cap;
	medge[k].flag= ME_EDGEDRAW;
	ls_error(LS_DEBUG, "k %d\n", k);

	dlm->totvert += 1;
	dlm->mvert= mvert;
}

static void make_flower(LGeneratorMesh *lgen, DispListMesh *dlm, TurtleState *ts)
{
	MEdge *medge;
	MFace *mface;
	Mesh *flower;

	flower= lgen->flower;
	if(!flower) {
		return;
	}
	merge_and_transform(dlm, flower, ts, medge, mface);
	dlm->totvert += flower->totvert;
}

/* callbacks */
static void mesh_initData(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;
	DispListMesh *dlm;
	int i, j;
	MEdge *medge;
	int num_edges;

	ts= default_turtle_state(ts);
	if (lgen->dmap_flag & LSYS_DISTRIBUTION_MAP) {
		memcpy(ts->m, lgen->dmap_m, sizeof(float)*16);
	}
	if (!lgen->dlm) {
		dlm= (DispListMesh *) MEM_callocN(sizeof(DispListMesh), "mesh_initData dlm");
		dlm->dontFreeVerts = dlm->dontFreeOther = dlm->dontFreeNors = 0;
		lgen->dlm= dlm;
	}
	lgen->gen.flag |= eGeneratorFlag_Initialized;
	lgen->prev_seg= -1;
	if(lgen->contour) {
		/* selected edges are copied in an extra array */
		lgen->num_sel_edges= 0;
		medge= lgen->contour->medge;
		num_edges= lgen->contour->totedge;
		lgen->sel_edges= MEM_mallocN(sizeof(MEdge)*num_edges, "sel edges");
		for(i= 0, j= 0; i<num_edges; i++, medge++) {
			if(medge->flag & 1) {
				memcpy(&lgen->sel_edges[j++], medge, sizeof(MEdge));
				lgen->num_sel_edges++;
			}
		}
	}
	/* root */
	lgen->mati= ts->mati;
	add_segment(lgen, lgen->dlm, lgen->contour, ts);

	lgen->state_flag &= ~LSYS_POP;
}

static void mesh_freeData(LGenerator *gen)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

	if(lgen->sel_edges) {
		MEM_freeN(lgen->sel_edges);
	}

}

static void mesh_cmd_popTurtleState(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

	lgen->state_flag &= LSYS_POP;
	lgen->prev_seg= ts->v.integer;
}

static void mesh_cmd_popTurtleState_pre(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;
	float pos[3];

	/* create caps at the end of a branch */
//	Mat4Translate(ts->m, 0.0, 0.0, ts->len);
//	memcpy(&pos, ts->m[3], 3*sizeof(float) );
//	make_cap(lgen->dlm, pos, lgen->lprof->npnts);
	make_flower(lgen, lgen->dlm, ts);
	lgen->mati= ts->mati;
	add_segment(lgen, lgen->dlm, lgen->contour, ts);
}


static void mesh_cmd_pushTurtleState(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

	ts->v.integer= lgen->prev_seg;
}

static void mesh_cmd_F(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

	lgen->mati= ts->mati;
	add_segment(lgen, lgen->dlm, lgen->contour, ts);

	lgen->state_flag &= ~LSYS_POP;
}

static void mesh_cmd_f(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

	lgen->prev_seg= -1;
}

static void mesh_cmd_I(LGenerator *gen, TurtleState *ts, Object *ob)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;
	Mesh *implant;
	DispListMesh *dlm;
	MEdge *medge;
	MFace *mface;

	if(ob->type!=OB_MESH) {
		ls_error(LS_WARNING, "Implant object %s does not have mesh data.\n", ob->id.name);
		return;
	}
	implant= (Mesh *) ob->data;
	dlm= lgen->dlm;
	merge_and_transform(dlm, implant, ts, medge, mface);
	dlm->totvert += implant->totvert;
}

static void mesh_cmd_curlyOpen(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

}

static void mesh_cmd_curlyClose(LGenerator *gen, TurtleState *ts)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;

}

/* Distribution Map */
/* Create an lsystem at every vertex */
DispListMesh *distribution_map_loop(Lsystem *ls, LGenerator *gen, DispListMesh *dmap)
{
	LGeneratorMesh *lgen= (LGeneratorMesh *) gen;
	int i;
	float vec[3], *q2, mat[3][3], tmat[4][4];

	for(i= 0; i < dmap->totvert; i++) {
		MTC_Mat4One(lgen->dmap_m);
		/* Copy position */
		memcpy(lgen->dmap_m[3], dmap->mvert[i].co, sizeof(float)*3);
		if (lgen->dmap_flag & LSYS_ORIENT_TO_NORMAL) {
			/* The tree grows into the direction of the normal */
			vec[0]= -dmap->mvert[i].no[0];
			vec[1]= -dmap->mvert[i].no[1];
			vec[2]= -dmap->mvert[i].no[2];

			q2= vectoquat(vec, 2, 2);

			QuatToMat3(q2, mat);
			Mat4CpyMat4(tmat, lgen->dmap_m);
			Mat4MulMat43(lgen->dmap_m, tmat, mat);
		}
		/* The age of the tree depends on the weight painting */
			/* still missing */
		execute_generator(ls, (LGenerator *) lgen);
	}
	return(lgen->dlm);
}
/***************************/

void add_persistent_generator(Lsystem *ls, LGenerator *gen)
{
	LGenerator *g;

	g= ls->generators.first;
	while(g) {
		if (g->type==gen->type) {
			/* remove old generator of the same type */
			BLI_freelinkN(&ls->generators, g);
			break;
		}
		g= g->next;
	}
	BLI_addtail(&ls->generators, gen);
}

void rem_persistent_generator(Lsystem *ls, eGeneratorType gtype)
{
	LGenerator *g;

	g= ls->generators.first;
	while(g) {
		if (g->type==gtype) {
			BLI_freelinkN(&ls->generators, g);
			break;
		}
		g= g->next;
	}
}

LGenerator *fetch_persistent_generator(Lsystem *ls, eGeneratorType gtype)
{
	LGenerator *g;

	g= ls->generators.first;
	while(g) {
		if (g->type==gtype) {
			return(g);
		}
	}
	return(NULL);
}

void initGenerator_callbacks(LGenerator *lg, eGeneratorType gtype)
{
	lg->flag= 0;
	lg->cmd_popTurtleState_pre= NULL;
	switch(gtype) {
	case eGeneratorType_Skeleton:
		lg->flag |= eGeneratorFlag_Persistent;
		lg->initData= skeleton_initData;
		lg->freeData= skeleton_freeData;
		lg->cmd_popTurtleState= skeleton_cmd_popTurtleState;
		lg->cmd_pushTurtleState= skeleton_cmd_pushTurtleState;
		lg->cmd_F= skeleton_cmd_F;
		lg->cmd_f= skeleton_cmd_f;
		lg->cmd_I= skeleton_cmd_I;
		lg->cmd_curlyOpen= skeleton_cmd_curlyOpen;
		lg->cmd_curlyClose= skeleton_cmd_curlyClose;
		/* because skeleton is a persistent generator */
		lg->draw= skeleton_draw;
		break;
	case eGeneratorType_Mesh:
		lg->initData= mesh_initData;
		lg->freeData= mesh_freeData;
		lg->cmd_popTurtleState_pre= mesh_cmd_popTurtleState_pre;
		lg->cmd_popTurtleState= mesh_cmd_popTurtleState;
		lg->cmd_pushTurtleState= mesh_cmd_pushTurtleState;
		lg->cmd_F= mesh_cmd_F;
		lg->cmd_f= mesh_cmd_f;
		lg->cmd_I= mesh_cmd_I;
		lg->cmd_curlyOpen= mesh_cmd_curlyOpen;
		lg->cmd_curlyClose= mesh_cmd_curlyClose;
		break;
	default:
		ls_error(LS_ERROR, "uncaught case in switch statement (initGenerator_callbacks)\n");
	}

}

/* Lsystem Core */
void execute_generator(Lsystem *ls, LGenerator *gen)
{
	LChar *l;
	int plen;
	float rand[3];
	float param;			/* needed by macro GROWING_FRAC */
	float *precalc_growfrac;	/* "" */
	int i;

	short error_flag= 0;
	/* reset turtle state */
	ls->defcolor[3]= 1.0;
	MCOLOR(ls->curts.mcol, ls->defcolor[0], ls->defcolor[1], ls->defcolor[2], ls->defcolor[3]);
	ls->curts.angle= ls->defangle;
	ls->curts.len= ls->deflen;
	/* generator may overwrite the default colour, angle and length */
	gen->initData(gen, &ls->curts);

	float x_axis[3]= {1.0, 0.0, 0.0};
	float y_axis[3]= {0.0, 1.0, 0.0};
	float z_axis[3]= {0.0, 0.0, 1.0};

	/* Precalculate ipos for generation time */
	if (ls->flag & LSYS_GROWING) {
		precalc_growfrac= MEM_mallocN(sizeof(float) * (ls->iter + 1), "precalc growfrac");
		for(i= 0; i<= ls->iter; i++) {
			precalc_growfrac[i]= calc_growfrac(ls, ls->gentime - (float) i);
		}
	}

	l= ls->prod;
	plen= ls->plen;

	while(plen--) {
		switch(l->c) {
		case '[':
			gen->cmd_pushTurtleState(gen, &ls->curts);
			push_turtle_state(ls);
			break;
		case ']':
			if (gen->cmd_popTurtleState_pre) {
				if (ls->flag & LSYS_GROWING) {
					ls->curts.sx *= precalc_growfrac[l->gen];
					ls->curts.sy *= precalc_growfrac[l->gen];
					ls->curts.sz *= precalc_growfrac[l->gen];
					gen->cmd_popTurtleState_pre(gen, &ls->curts);
				}
			}
			pop_turtle_state(ls);
			gen->cmd_popTurtleState(gen, &ls->curts);
			break;
		case '{':
			error_flag |= LSYS_MISSING_CURLY;
			gen->cmd_curlyOpen(gen, &ls->curts);
			break;
		case '}':
			error_flag &= ~LSYS_MISSING_CURLY;
			gen->cmd_curlyClose(gen, &ls->curts);
			break;
		case 'F':
			CHECK_USEDEF_F1(l, ls->curts.len);
			GROWING_FRAC(l, LSYS_GROWLEN);
			/* move into z-direction and draw */
			Mat4Translate(ls->curts.m, 0.0, 0.0, param);
			if (ls->flag & LSYS_LEVODET) {
				ls->curts.acc_F += param;
				if (ls->curts.acc_F >= ls->lod_minsize) {
					ls->curts.acc_F= 0.0;
				} else {
					break;
				}
			}
			gen->cmd_F(gen, &ls->curts);
			break;
		case 'f':
			CHECK_USEDEF_F1(l, ls->curts.len);
			GROWING_FRAC(l, LSYS_GROWLEN);
			/* move into y-direction */

			Mat4Translate(ls->curts.m, 0.0, 0.0, param);
			if (ls->flag & LSYS_LEVODET) {
				ls->curts.acc_f += param;
				if (ls->curts.acc_f >= ls->lod_minsize) {
					ls->curts.acc_f= 0.0;
				} else {
					break;
				}
			}
			gen->cmd_f(gen, &ls->curts);
			break;
		case 'I':
			if(l->idp) {
				gen->cmd_I(gen, &ls->curts, (Object *) l->idp);
			}
			break;
		/* Generator independent part */
		case 'c':
			/* A c() command without parameters resets the colour to the default colour. */
			CHECK_USEDEF_F4(l, ls->defcolor[0], ls->defcolor[1], ls->defcolor[2], ls->defcolor[3]);
			MCOLOR(ls->curts.mcol, l->param[0], l->param[1], l->param[2], l->param[3]);
			break;
		case 'm':
			CHECK_USEDEF_F1(l, 1);
			ls->curts.mati= l->param[0];
			break;
		case '+':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(-param), x_axis);
			break;
		case '-':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(param), x_axis);
			break;
		case '<':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(-param), z_axis);
			break;
		case '>':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(param), z_axis);
			break;
		case '&':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(-param), y_axis);
			break;
		case '^':
			CHECK_USEDEF_F1(l, ls->curts.angle);
			GROWING_FRAC(l, LSYS_GROWANG);
			Mat4Rotate(ls->curts.m, RADIANS(param), y_axis);
			break;
		case '~':
			GROWING_FRAC(l, LSYS_GROWANG);
			/* to make sure we get the same random value for every rebuilt
			   we use the current position */
			BLI_srand( ls->curts.m[3][0] + ls->curts.m[3][1] + ls->curts.m[3][2]  );
			rand[0]= BLI_frand();
			rand[1]= BLI_frand();
			rand[2]= BLI_frand();
			Mat4Rotate(ls->curts.m, param, rand);
			break;
		case 't':			/* not yet implemented */
			break;
		case '"':
			CHECK_USEDEF_F1(l, 1.0 + ls->len_step);
			ls->curts.len *= l->param[0];
			break;
		case '\'':
			CHECK_USEDEF_F1(l, 1.0 - ls->len_step);
			ls->curts.len *= l->param[0];
			break;
		case ';':
			CHECK_USEDEF_F1(l, 1.0 + ls->angle_step);
			ls->curts.angle *= l->param[0];
			break;
		case ':':
			CHECK_USEDEF_F1(l, 1.0 - ls->angle_step);
			ls->curts.angle *= l->param[0];
			break;
		case 's':
			CHECK_USEDEF_F3(l, 0.9, 0.9, 0.9);
			GROWING_FRAC_F3(l, LSYS_GROWSIZE);
			ls->curts.sx= l->param[0];
			ls->curts.sy= l->param[1];
			ls->curts.sz= l->param[2];
			break;
		default:
			ls_error(LS_WARNING, "-------> %d\n", l->c);
			error_flag |= LSYS_UNRES_RULE;
		};
		l++;
		//debug_lsystem(ls, TURTLE_STACK);
	}
	if (ls->flag & LSYS_GROWING)
		MEM_freeN(precalc_growfrac);

	if (error_flag) {
		if(error_flag & LSYS_UNRES_RULE)
			ls_error(LS_WARNING, "At least one unresolved rule in production string. Make sure all rules can be resolved to some drawing commands. If the unresolved rule is at the end of a replacement string, this might be correct and allows for recursion, though.\n");
		if(error_flag & LSYS_MISSING_CURLY)
			ls_error(LS_WARNING, "Curly bracket not closed before end of production string.\n");
	}
	//debug_lsystem(ls, LVERTS);

	if ( (gen->flag & eGeneratorFlag_Persistent) == 0) {
		gen->freeData(gen);
	}
}

